
// Writes contents of fifo to USB Keyboard one key by each other and wait the specified time after each character
// A character with value 0 means: no key but just wait
// Special Keys (variable special) mean: only set but no autorelease - there is a special Special Key´for that...
unsigned long writebuffer(int index) {
  mj_keybuffer k;
  unsigned long del;
  int i;
  int special;
  tasklist[index].active = true;
  if ( mj_fifo_stored() > 0 ) {
    k = mj_fifo_getelement();
    i = k.d & 0x0f;
    if ((k.d & 0x80) > 0) {
      special = 1;
    } else {
      special = 0;
    }
    if (i >= 0 && i <= 9) {
      del = mj_keydelay[i];
    } else {
      del = mj_keydelay[0];
    }
    if (special == 0) {
      if (k.c != KEY_NONE) {
        Keyboard.write(k.c);
      }
    } else {
      if (k.c != KEY_RELEASE) {
        Keyboard.press(k.c);
      } else {
        Keyboard.releaseAll();
      }
    }
    return (del);
  }
  return (5);
}

// Process Actions Queued in the Action Fifo
unsigned long processactions(int index) {
  static const char win_shutdown[] PROGMEM = "shutdown -s -t 10 -f\\d5\n\\d1\nexit\\d7\n";
  static const char win_shutdown_alt[] PROGMEM = "\\d1\\s82\\s80\\d5\\sd4\\sff\\d2\\s81\\sb3\\d5\\sff\\d2\n\\d3\\Xd9\\d7\\Xd9\n";
  static const char win_lock_screen[] PROGMEM = "\\d1\\s83l\\d3\\sff";
  static const char win_explorer[] PROGMEM = "\\d1\\s83e\\d3\\sff";
  static const char win_run[] PROGMEM = "\\d1\\s83r\\d3\\sff";
  static const char win_cmd[] PROGMEM = "\\d1cmd\\d3\n\\d1";
  byte action;
  char nothing[] = "";
  tasklist[index].active = true;
  if ( mj_fifo_a_stored() > 0 ) {
    action = mj_fifo_a_getelement_n(); // Just take a look at coming action but not removing from fifo yet
    if (action > 0 && action < 64) { // Actions which send something to keyboard are only processed if keyboard-fifo is empty
      if (mj_fifo_stored() != 0) {
        return (50);
      }
    }
    action = mj_fifo_a_getelement(); // Get next action from fifo
    if (action > 0) {                                     // 0    -> Do Nothing
      if (action >= 1 && action <= 14) {                  // 1-14 -> Type string from eeprom one of 1-14
        putfifo_string(nothing, action);
      } else if (action > 14 && action < 32) {            // 15-31 -> Type strings from eeprom - combining
        switch (action) {  // This actiona are recursive
          case 15:         // They queue known actions    // 15 -> Type Strings 1 and 2 from EEPROM
            mj_fifo_a_putelement(1);
            mj_fifo_a_putelement(2);
            break;
          case 16:                                        // 16 -> Type Strings 3 and 4 from EEPROM
            mj_fifo_a_putelement(3);
            mj_fifo_a_putelement(4);
            break;
          case 17:                                        // 17 -> Type Strings 5 and 6 from EEPROM
            mj_fifo_a_putelement(5);
            mj_fifo_a_putelement(6);
            break;
          case 18:                                        // 18 -> Type Strings 7 and 8 from EEPROM
            mj_fifo_a_putelement(7);
            mj_fifo_a_putelement(8);
            break;
          case 19:                                        // 19 -> Type Strings 9 and 10 from EEPROM
            mj_fifo_a_putelement(9);
            mj_fifo_a_putelement(10);
            break;
          case 20:                                        // 20 -> Type Strings 11 and 12 from EEPROM
            mj_fifo_a_putelement(11);
            mj_fifo_a_putelement(12);
            break;
          case 21:                                        // 21 -> Type Strings 13 and 14 from EEPROM
            mj_fifo_a_putelement(13);
            mj_fifo_a_putelement(14);
            break;
          case 22:                                        // 22 -> Type Strings 1, 2 and 3 from EEPROM
            mj_fifo_a_putelement(15);
            mj_fifo_a_putelement(3);
            break;
          case 23:                                        // 23 -> Type Strings 4, 5 and 6 from EEPROM
            mj_fifo_a_putelement(4);
            mj_fifo_a_putelement(17);
            break;
          case 24:                                        // 24 -> Type Strings 7, 8 and 9 from EEPROM
            mj_fifo_a_putelement(18);
            mj_fifo_a_putelement(9);
            break;
          case 25:                                        // 25 -> Type Strings 10, 11 and 12 from EEPROM
            mj_fifo_a_putelement(10);
            mj_fifo_a_putelement(20);
            break;
          case 26:                                        // 26 -> Type Strings 1, 2, 3, 4 from EEPROM
            mj_fifo_a_putelement(15);
            mj_fifo_a_putelement(16);
            break;
          case 27:                                        // 27 -> Type Strings 5, 6, 7, 8 from EEPROM
            mj_fifo_a_putelement(17);
            mj_fifo_a_putelement(18);
            break;
          case 28:                                        // 28 -> Type Strings 9, 10, 11, 12 from EEPROM
            mj_fifo_a_putelement(19);
            mj_fifo_a_putelement(20);
            break;
          case 29:                                        // 29 -> Type Strings 1, 2, 3, 4, 5, 6, 7, 8 from EEPROM
            mj_fifo_a_putelement(22);
            mj_fifo_a_putelement(23);
            break;
          case 30:                                        // 30 -> Type Strings 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 from EEPROM
            mj_fifo_a_putelement(25);
            mj_fifo_a_putelement(24);
            break;
          case 31:                                        // 31 -> Type all Strings (1 - 14) from EEPROM
            mj_fifo_a_putelement(30);
            mj_fifo_a_putelement(21);
            break;
        }
      } else if (action >= 32 && action < 64) { // Predefined Keyboard Actions
        switch (action) {
          case 32:                                        // 32 -> Shutdown Windows Method #1
            putfifo_string(win_run, 0xff);  // String from PROGMEM = 0xff
            putfifo_string(win_cmd, 0xff);
            putfifo_string(win_shutdown, 0xff);
            break;
          case 33:                                        // 33 -> Shutdown Windows Method #2
            putfifo_string(win_shutdown_alt, 0xff);
            break;
          case 34:                                        // 34 -> Lock Windows Screen (WIN+L)
            putfifo_string(win_lock_screen, 0xff);
            break;
          case 35:                                        // 35 -> Start Windows Explorer (WIN+E)
            putfifo_string(win_explorer, 0xff);
            break;
          case 36:                                        // 36 -> Run Application (WIN+R)
            putfifo_string(win_run, 0xff);
            break;
          case 37:                                        // 37 -> Run CMD
            putfifo_string(win_run, 0xff);
            putfifo_string(win_cmd, 0xff);
            break;
        }
      } else if (action > 64 && action < 128) { // Mouse...
        switch (action) {
          case 64:                                      // 64 -> Move Mouse 1 Pixel Up
            Mouse.move(0, -1);
            break;
          case 65:                                      // 65 -> Move Mouse 1 Pixel Down
            Mouse.move(0, 1);
            break;
          case 66:                                      // 66 -> Move Mouse 1 Pixel Left
            Mouse.move(-1, 0);
            break;
          case 67:                                      // 67 -> Move Mouse 1 Pixel Right
            Mouse.move(0, -1);
            break;
          case 68:                                      // 68 -> Move Mouse 1 Pixel Up+Left
            Mouse.move(-1, -1);
            break;
          case 69:                                      // 69 -> Move Mouse 1 Pixel Up+Right
            Mouse.move(1, -1);
            break;
          case 70:                                      // 70 -> Move Mouse 1 Pixel Down+Left
            Mouse.move(-1, 1);
            break;
          case 71:                                      // 71 -> Move Mouse 1 Pixel Down+Right
            Mouse.move(1, 1);
            break;
          case 72:                                      // 72 -> Move Mouse 10 Pixel Up
            Mouse.move(0, -10);
            break;
          case 73:                                      // 73 -> Move Mouse 10 Pixel Down
            Mouse.move(0, 10);
            break;
          case 74:                                      // 74 -> Move Mouse 10 Pixel Left
            Mouse.move(-10, 0);
            break;
          case 75:                                      // 75 -> Move Mouse 10 Pixel Right
            Mouse.move(0, -10);
            break;
          case 76:                                      // 76 -> Move Mouse 10 Pixel Up+Left
            Mouse.move(-10, -10);
            break;
          case 77:                                      // 77 -> Move Mouse 10 Pixel Up+Right
            Mouse.move(10, -10);
            break;
          case 78:                                      // 78 -> Move Mouse 10 Pixel Down+Left
            Mouse.move(-10, 10);
            break;
          case 79:                                      // 79 -> Move Mouse 10 Pixel Down+Right
            Mouse.move(10, 10);
            break;
          case 96:                                      // 96 -> Press left Mouse Button
            Mouse.press(MOUSE_LEFT);
            break;  
          case 97:                                      // 97 -> Release left Mouse Button
            Mouse.release(MOUSE_LEFT);
            break;  
          case 98:                                      // 98 -> Press right Mouse Button
            Mouse.press(MOUSE_RIGHT);
            break;  
          case 99:                                      // 99 -> Release right Mouse Button
            Mouse.release(MOUSE_RIGHT);
            break;  
        }
      } else if (action >= 250 && action <= 252 ) {
        switch (action) {
          case 250:                                       // 250 -> Disable Jiggler
            mj_jiggler_enable = 0;
            mj_ledtiming_changed = true;
            break;
          case 251:                                       // 251 -> Enable Jiggler
            mj_jiggler_enable = 1;
            mj_ledtiming_changed = true;
            break;
          case 252:                                       // 252 -> Toggle Jiggler State
            if (mj_jiggler_enable == 0) {
              mj_jiggler_enable = 1;
            } else {
              mj_jiggler_enable = 0;
            }
            mj_ledtiming_changed = true;
            break;
        }
      }
    }
  }
  return (20);
}

void writekeypressed(int num, int pressed) {
  Serial.print(F("Key"));
  Serial.print(num);
  Serial.print(F(" was "));
  Serial.print(pressed ? "relea" : "pres");
  Serial.println(F("sed"));
}

unsigned long buttons(int index) {
  static byte laststate[9] = { HIGH, HIGH, HIGH, HIGH, HIGH, HIGH, HIGH, HIGH, HIGH };
  byte state;
  int a;
  long int ms;
  tasklist[index].active = true;
  for (a = 0; a < 9; a++) {
    if (mj_keypin[a] != 0xff) {
      state = (digitalRead(mj_keypin[a])) ? HIGH : LOW;
      if (state == LOW && laststate[a] == HIGH) {
        writekeypressed(a + 1, 0);
        if (mj_action[a][0][0] != 0) {
          mj_fifo_a_putelement(mj_action[a][0][0]);
        }
        if (mj_action[a][0][1] != 0) {
          mj_fifo_a_putelement(mj_action[a][0][1]);
        }
      }
      if (state == HIGH && laststate[a] == LOW) {
        writekeypressed(a + 1, 1);
        if (mj_action[a][1][0] != 0) {
          mj_fifo_a_putelement(mj_action[a][1][0]);
        }
        if (mj_action[a][1][1] != 0) {
          mj_fifo_a_putelement(mj_action[a][1][1]);
        }
      }
      laststate[a] = state;
    }
  }
  return (20);
}

unsigned long autotype(int index) {
  static int finished = false;
  static int initialized = false;
  if (initialized == false) {
    initialized = true;
    if (mj_autotype_1 > 0) {
      return (mj_autotype_delay);
    } else { // Disable task if no autotyping is used
      tasklist[index].active = false;
      tasklist[index].used = false;
      return (0);
    }
  }
  if (finished == false) {
    mj_fifo_a_putelement(mj_autotype_1);
    if (mj_autotype_2 > 0) {
      mj_fifo_a_putelement(mj_autotype_2);
    }
    finished = true;
    tasklist[index].active = false;
    tasklist[index].used = false;
  }
  return (0);
}

// Led Blinker
// mj_ledmode=0 statically blink
// mj_ledmode=1 heartbeat blink
// mj_ledmode=2 see 0 - but inverted timing jiggler enabled/disabled
// mj_ledmode=3 see 1 - but inverted timing jiggler enabled/disabled
// mj_ledmode=4 statically fade in/out
// mj_ledmode=5 heartbeat fade in/out
// mj_ledmode=6 see 4 - but inverted timing jiggler enabled/disabled
// mj_ledmode=7 see 5 - but inverted timing jiggler enabled/disabled
// mj_ledmode>7 see 5/7 (Bit 1 tells inverting jiggler enable/disable slowdown)
// mj_ledoption bit 0 = invert output led1
// mj_ledoption bit 1 = invert output led2

unsigned long led(int index) {
  static int state = 0;
  static unsigned long t[4] = {
    0, 0, 0, 0
  };
  static int value;
  int i = 0;
  unsigned int sleep = 500;
  tasklist[index].active = true;
  if (mj_ledtiming_changed == true) {
    // As this calculation uses division and multiplication we only calculate on change
    mj_ledtiming_changed = false;
    switch (mj_ledmode) {
      case 0:
      case 2:
        t[0] = (500 * (unsigned long) mj_ledtiming[0]) / 10000;
        t[1] = (500 * (unsigned long) mj_ledtiming[1]) / 10000;
        break;
      case 1:
      case 3:
        t[0] = (166 * (unsigned long) mj_ledtiming[0]) / 10000;
        t[1] = (166 * (unsigned long) mj_ledtiming[1]) / 10000;
        t[2] = (500 * (unsigned long) mj_ledtiming[2]) / 10000;
        break;
      case 4:
      case 6:
        t[0] = (31 * (unsigned long) mj_ledtiming[0]) / 10000;
        t[1] = (31 * (unsigned long) mj_ledtiming[1]) / 10000;
        t[2] = (500 * (unsigned long) mj_ledtiming[2]) / 10000;
        break;
      case 5:
      case 7:
      default:
        t[0] = (10 * (unsigned long) mj_ledtiming[0]) / 10000;
        t[1] = (10 * (unsigned long) mj_ledtiming[1]) / 10000;
        t[2] = (333 * (unsigned long) mj_ledtiming[2]) / 10000;
        break;
    }
    if ((mj_ledmode & 0x02) > 0) {
      i = !mj_jiggler_enable;
    } else {
      i = mj_jiggler_enable;
    }
    if (i == 0) {
      t[0] *= 3;
      t[1] *= 3;
      t[2] *= 3;
      t[3] *= 3;
    }
  }
  switch (mj_ledmode) {
    case 0:
    case 2:
      switch (state) {
        case 0:
          value = 255;
          sleep = t[0];
          state = 1;
          break;
        default:
          value = 0;
          sleep = t[1];
          state = 0;
          break;
      }
      break;
    case 1:
    case 3:
      switch (state) {
        case 0:
          value = 255;
          sleep = t[0];
          state = 1;
          break;
        case 1:
          value = 0;
          sleep = t[1];
          state = 2;
          break;
        case 2:
          value = 255;
          sleep = t[0];
          state = 3;
          break;
        default:
          value = 0;
          sleep = t[2];
          state = 0;
          break;
      }
      break;
    case 4:
    case 6:
      switch (state) {
        case 0:
          value += 16;
          sleep = t[0];
          if (value > 255) {
            value = 255;
            state = 1;
          }
          break;
        case 1:
          value -= 16;
          sleep = t[1];
          if (value < 0) {
            value = 0;
            state = 2;
          }
          break;
        case 2:
        default:
          sleep = t[2];
          state = 0;
          break;
      }
      break;
    case 5:
    case 7:
    default:
      switch (state) {
        case 0:
          value += 16;
          sleep = t[0];
          if (value > 255) {
            value = 255;
            state = 1;
          }
          break;
        case 1:
          value -= 16;
          sleep = t[1];
          if (value < 0) {
            value = 0;
            state = 2;
          }
          break;
        case 2:
          value += 16;
          sleep = t[0];
          if (value > 255) {
            value = 255;
            state = 3;
          }
          break;
        case 3:
          value -= 16;
          sleep = t[1];
          if (value < 0) {
            value = 0;
            state = 4;
          }
          break;
        default:
          sleep = t[2];
          state = 0;
          break;
      }
      break;
  }
  if (mj_ledpin1 != 0xff) {
    if ((mj_ledoption & 0x0001) == 0) {
      if (digitalPinHasPWM(mj_ledpin1) == true) {
        analogWrite(mj_ledpin1, value);
      } else {
        digitalWrite(mj_ledpin1, value > 127 ? 1 : 0);
      }
    } else {
      if (digitalPinHasPWM(mj_ledpin1) == true) {
        analogWrite(mj_ledpin1, 255 - value);
      } else {
        digitalWrite(mj_ledpin1, value > 127 ? 0 : 1);
      }
    }
  }
  if (mj_ledpin2 != 0xff) {
    if ((mj_ledoption & 0x0002) == 0) {
      if (digitalPinHasPWM(mj_ledpin2) == true) {
        analogWrite(mj_ledpin2, value);
      } else {
        digitalWrite(mj_ledpin2, value > 127 ? 1 : 0);
      }
    } else {
      if (digitalPinHasPWM(mj_ledpin2) == true) {
        analogWrite(mj_ledpin2, 255 - value);
      } else {
        digitalWrite(mj_ledpin2, value > 127 ? 0 : 1);
      }
    }
  }
  return (sleep);
}

unsigned long jiggler(int index) {
  unsigned long wait = 0;
  static int x, y, mx, my, state = 0;
  tasklist[index].active = true;
  switch (state) {
    case 0:
      if (mj_jiggler_enable == 1) {
        //random-generatre move, initial move
        x = random(0 - mj_jiggler_pixel, mj_jiggler_pixel);
        y = random(0 - mj_jiggler_pixel, mj_jiggler_pixel);
        mx = 0 - x;
        my = 0 - y;
        Mouse.move(x, y, 0);
        wait = random(50, 150);
        state = 1;
      } else {
        wait = 500;
      }
      break;
    case 1:
      // move backwards;
      Mouse.move(mx, my, 0);
      wait = random(50, 150);
      state = 2;
      break;
    case 2:
      // move backwards again;
      Mouse.move(mx, my, 0);
      wait = random(50, 150);
      state = 3;
      break;
    case 3:
      // move to original position;
      Mouse.move(x, y, 0);
      wait = random(mj_jiggler_min, mj_jiggler_max);
      state = 0;
      break;
    default:
      state = 0;
      break;
  }
  return (wait);
}

// Reads from Serial and writes into buffer. Calls parsestring() when \r or \n comes and deletes last char if 0x7f comes (Backspace)
unsigned long readser(int index) {
  unsigned char byte;
  static char buffer[SERIAL_BUFFERSIZE];
  static unsigned int numbytes = 0;
  int i = 10;
  tasklist[index].active = true;
  if (Serial.available() > 0) {
    i = 0;
    byte = (unsigned char) Serial.read();
    if (byte == '\n' || byte == '\r') {
      Serial.print(F("\n\r"));
      if (numbytes > 0) {
        buffer[numbytes] = '\0';
        parsestring(buffer, numbytes);
        numbytes = 0;
      }
    } else if (byte == 0x7f || byte == 0x08) {  // Del or Backspace
      if (numbytes > 0) {
        numbytes--;
        Serial.write(byte); // Go one step back - using the same code as the serial client used (0x08 or 0x7f)
        Serial.write(' ');  // Overwrite old char with space
        Serial.write(byte); // Go one step back again
      }
    } else if (byte > 0x7f || byte < 0x20) { // Non-ASCII and non-printable will be ignored
      ;
    } else if (numbytes < (SERIAL_BUFFERSIZE - 1)) {
      buffer[numbytes] = byte;
      numbytes++;
      Serial.write(byte);
    }
  }
  return ((unsigned long) i);
}
